from flask_nemo.query.interface import SimpleQuery
from flask_nemo.query.annotation import AnnotationResource
from flask_nemo.query.resolve import Resolver, LocalRetriever
from flask_nemo import Nemo
from unittest import TestCase
from flask import Flask
from MyCapytain.common.reference import URN
from MyCapytain.resolvers.cts.local import CTSCapitainsLocalResolver
from werkzeug.exceptions import NotFound
import logging


def dict_list(l):
    return [(a.uri, a.target.objectId, a.target.subreference, a.type_uri) for a in l]


class TestSimpleQuery(TestCase):
    """ Test simple query interface """
    def setUp(self):
        self.resolver = Resolver(LocalRetriever(path="./tests/test_data/"))
        self.one = (URN("urn:cts:latinLit:phi1294.phi002.perseus-lat2:6.1"), "interface/treebanks/treebank1.xml", "dc:treebank")
        self.two = (URN("urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.5"), "interface/treebanks/treebank2.xml", "dc:treebank")
        self.three = (URN("urn:cts:latinLit:phi1294.phi002.perseus-lat2:6.1"), "interface/images/N0060308_TIFF_145_145.tif", "dc:image")
        self.four = AnnotationResource(
            "interface/researchobject/researchobject.json",
            type_uri="dc:researchobject",
            target=URN("urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.pr.1"),
            resolver=self.resolver
        )
        self.one_anno = AnnotationResource(
            self.one[1],
            type_uri=self.one[2],
            target=self.one[0],
            resolver=self.resolver
        )
        self.two_anno = AnnotationResource(
            self.two[1],
            type_uri=self.two[2],
            target=self.two[0],
            resolver=self.resolver
        )
        self.three_anno = AnnotationResource(
            self.three[1],
            type_uri=self.three[2],
            target=self.three[0],
            resolver=self.resolver
        )
        self.fourth_anno = AnnotationResource(
            self.three[1],
            type_uri=self.three[2],
            target=("urn:cts:latinLit:phi1294.phi002.perseus-lat2", "1-2"),
            resolver=self.resolver
        )
        self.app = Flask("app")
        logger = logging.getLogger('my-logger')
        logger.propagate = False
        self.nautilus = CTSCapitainsLocalResolver(["tests/test_data/interface/latinLit"], logger=logger)
        self.nemo = Nemo(app=self.app, resolver=self.nautilus, base_url="/")
        self.query = SimpleQuery(
            [
                self.one,
                self.two,
                self.three,
                self.four
            ],  # List of annotations
            self.resolver
        )
        self.query.process(self.nemo)

    def test_process(self):
        """ Check that all annotations are taken care of"""
        self.assertEqual(len(self.query.annotations), 4, "There should be 4 annotations")

    def test_get_all(self):
        """ Check that get all returns 3 annotations """
        hits, annotations = self.query.getAnnotations(None)
        self.assertEqual(hits, 4, "There should be 4 annotations")
        self.assertCountEqual(dict_list(annotations), dict_list(
            [
                self.one_anno,
                self.two_anno,
                self.three_anno,
                self.four
            ]), "There should be all three annotation")

    def test_get_exact_match(self):
        """ Ensure exact match works """
        # Higher level annotation or same match
        hits, annotations = self.query.getAnnotations(("urn:cts:latinLit:phi1294.phi002.perseus-lat2", "6.1"))
        self.assertEqual(hits, 2, "There should be 2 annotations")
        self.assertCountEqual(
            [anno.uri for anno in annotations],
            [
                "interface/images/N0060308_TIFF_145_145.tif",
                "interface/treebanks/treebank1.xml"
            ],
            "Only 2 annotation s having 6.1 should match"
        )

        # Deep Annotations
        hits, annotations = self.query.getAnnotations(("urn:cts:latinLit:phi1294.phi002.perseus-lat2", "1.pr.1"))
        self.assertEqual(annotations[0], self.four, "Deepest node should match as well")

        # It should take URNs as well
        hits, annotations = self.query.getAnnotations(URN("urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.pr.1"))
        self.assertEqual(annotations[0], self.four, "Deepest node should match as well")

    def test_get_range_match(self):
        """ Ensure range match as well """
        # Higher level annotation or same match
        hits, annotations = self.query.getAnnotations(("urn:cts:latinLit:phi1294.phi002.perseus-lat2", "6.1-6.2"))
        self.assertEqual(hits, 2, "There should be 2 annotations")
        self.assertCountEqual(
            [anno.uri for anno in annotations],
            [
                "interface/images/N0060308_TIFF_145_145.tif",
                "interface/treebanks/treebank1.xml"
            ],
            "Only 2 annotation s having 6.1 should match"
        )

    def test_deep_first(self):
        # Deep Annotations
        hits, annotations = self.query.getAnnotations(URN("urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.pr.1-1.pr.3"))
        self.assertEqual(annotations[0], self.four, "Deepest node should match as well")
        # Test with tuple
        hits, annotations = self.query.getAnnotations(("urn:cts:latinLit:phi1294.phi002.perseus-lat2", "1.pr.1-1.pr.3"))
        self.assertEqual(annotations[0], self.four, "Deepest node should match as well")

    def test_get_resource(self):
        resource = self.query.getResource("0f9a85344190c3a0376f67764f7e193ffb175c1b59fefb0017c15a5cd8baaa33")
        self.assertEqual(resource, self.four, "Getting one resource should work")

        with self.assertRaises(NotFound, msg="Getting a resource for an unknown sha should raise NotFound from werkzeug"):
            self.query.getResource("sasfd")



"""
    Test for routes functions : ensure responses are correct with mocked call to API
"""

from tests.test_resources import NemoResource
from flask_nemo import Nemo
from flask import Flask
from mock import Mock, patch, call
from jinja2.exceptions import TemplateNotFound


class NemoTestRoutesBusiness(NemoResource):
    """ Test Suite for Nemo
    """
    def test_route_index(self):
        """ Check that index return the template
        """
        self.assertEqual(self.nemo.r_index(), {"template": "main::index.html"})

    def test_route_assets_404(self):
        with patch('flask_nemo.abort') as abort:
            self.nemo.r_assets("js", "wrong-js.js")
            abort.assert_called_once_with(404)

    def test_route_assets_all(self):
        nemo = Nemo(
            statics=["testing_data/getcapabilities.xml"],
            js=["testing_data/getcapabilities.xml"],
            css=["testing_data/style.css"]
        )
        nemo.blueprint = Mock()
        nemo.register_assets()
        with patch("flask_nemo.send_from_directory") as patched:
            nemo.r_assets("js", "getcapabilities.xml")
            nemo.r_assets("static", "getcapabilities.xml")
            nemo.r_assets("css", "style.css")
            patched.assert_has_calls([
                call(directory="testing_data", filename="getcapabilities.xml"),
                call(directory="testing_data", filename="getcapabilities.xml"),
                call(directory="testing_data", filename="style.css")
            ])

    def test_register_route(self):
        app = Flask(__name__)
        nemo = Nemo(app=app, base_url="/perseus")
        nemo.register()
        self.assertIn("flask_nemo", app.blueprints)

        rules = [(rule.rule, rule.endpoint) for rule in app.url_map.iter_rules()]
        self.assertIn("/perseus/collections/<objectId>", [rule[0] for rule in rules])
        self.assertIn("flask_nemo.r_collection", [rule[1] for rule in rules])

        app = Flask(__name__)
        nemo = Nemo("nemo", app=app)
        nemo.register()
        self.assertIn("nemo", app.blueprints)

        rules = [(rule.rule, rule.endpoint) for rule in app.url_map.iter_rules()]
        self.assertIn("/nemo/collections/<objectId>", [rule[0] for rule in rules])
        self.assertIn("nemo.r_collection", [rule[1] for rule in rules])

        nemo = Nemo()
        self.assertEqual(nemo.register(), None)

    def test_additional_template(self):
        # Line 568-575
        app = Flask(__name__)
        nemo = Nemo(app=app, templates={"addendum": "tests/test_data/plugin_templates_main/main"})
        blueprint = nemo.create_blueprint()

        html, path, function = blueprint.jinja_loader.get_source("", "addendum::container.html")
        self.assertIn("I am A CONTAINER ! Isn't it sweet !", html)

        with self.assertRaises(TemplateNotFound):
            html, path, function = blueprint.jinja_loader.get_source("", "addendum::unknown.html")


# -*- coding: utf-8 -*-
from flask_nemo.plugin import PluginPrototype
from flask import jsonify, request, url_for, Response
import MyCapytain.common.reference


class AnnotationsApiPlugin(PluginPrototype):
    """AnnotationsApiPlugin adds routes to Nemo from which annotations can be retrieved

    This plugins contains two routes only registered at

    - /api/annotations/?target=<URN Target> which is the collection in which to search
    - /api/annotations/<SHA Identifier of the annotation> is the annotation object
    - /api/annotations/<SHA Identifier of the annotation>/body is the proxy for the annotation body

    The response are conform to https://www.w3.org/TR/annotation-model/#annotation-collection

    :param queryinterface: QueryInterface to use to retrieve annotations
    :type queryinterface: QueryInterface

    """

    JSONLD_CONTEXT = {
        "": "http://www.w3.org/ns/anno.jsonld",
        "dc": "http://purl.org/dc/terms/",
        "owl": "http://www.w3.org/2002/07/owl#",
        "nemo": "https://capitains.github.io/flask-capitains-nemo/ontology/#"
    }
    ROUTES = [
        ("/api/annotations", "r_annotations", ["GET"]),
        ("/api/annotations/<sha>", "r_annotation", ["GET"]),
        ("/api/annotations/<sha>/body", "r_annotation_body", ["GET"])
    ]

    def __init__(self, queryinterface, *args, **kwargs):
        super(AnnotationsApiPlugin, self).__init__(*args, **kwargs)
        self.__queryinterface__ = queryinterface

    # TODO we should have a response at the base of annotations/api that returns link types and link relations
    # showing the next level of options 
    
    def r_annotations(self):
        """ Route to retrieve annotations by target

        :param target_urn: The CTS URN for which to retrieve annotations  
        :type target_urn: str
        :return: a JSON string containing count and list of resources
        :rtype: {str: Any}
        """

        target = request.args.get("target", None)
        wildcard = request.args.get("wildcard", ".", type=str)
        include = request.args.get("include")
        exclude = request.args.get("exclude")
        limit = request.args.get("limit", None, type=int)
        start = request.args.get("start", 1, type=int)
        expand = request.args.get("expand", False, type=bool)

        if target:

            try:
                urn = MyCapytain.common.reference.URN(target)
            except ValueError:
                return "invalid urn", 400

            count, annotations = self.__queryinterface__.getAnnotations(urn, wildcard=wildcard, include=include,
                                                                        exclude=exclude, limit=limit, start=start,
                                                                        expand=expand)
        else:
            #  Note that this implementation is not done for too much annotations
            #  because we do not implement pagination here
            count, annotations = self.__queryinterface__.getAnnotations(None, limit=limit, start=start, expand=expand)
        mapped = []
        response = {
            "@context": type(self).JSONLD_CONTEXT,
            "id": url_for(".r_annotations", start=start, limit=limit),
            "type": "AnnotationCollection",
            "startIndex": start,
            "items": [
            ],
            "total": count
        }
        for a in annotations:
            mapped.append({
                "id": url_for(".r_annotation", sha=a.sha),
                "body": url_for(".r_annotation_body", sha=a.sha),
                "type": "Annotation",
                "target": a.target.to_json(),
                "dc:type": a.type_uri,
                "owl:sameAs": [a.uri],
                "nemo:slug": a.slug
            })
        response["items"] = mapped
        response = jsonify(response)
        return response

    def r_annotation(self, sha):
        """ Route to retrieve contents of an annotation resource

        :param uri: The uri of the annotation resource
        :type uri: str
        :return: annotation contents
        :rtype: {str: Any}
        """
        annotation = self.__queryinterface__.getResource(sha)
        if not annotation:
            return "invalid resource uri", 404
        response = {
            "@context": type(self).JSONLD_CONTEXT,
            "id": url_for(".r_annotation", sha=annotation.sha),
            "body": url_for(".r_annotation_body", sha=annotation.sha),
            "type": "Annotation",
            "target": annotation.target.to_json(),
            "owl:sameAs": [annotation.uri],
            "dc:type": annotation.type_uri,
            "nemo:slug": annotation.slug
        }
        return jsonify(response)

    def r_annotation_body(self, sha):
        """ Route to retrieve contents of an annotation resource

        :param uri: The uri of the annotation resource
        :type uri: str
        :return: annotation contents
        :rtype: {str: Any}
        """
        annotation = self.__queryinterface__.getResource(sha)
        if not annotation:
            return "invalid resource uri", 404
        # TODO this should inspect the annotation content
        # set appropriate Content-Type headers
        # and return the actual content
        content = annotation.read()
        if isinstance(content, Response):
            return content
        headers = {"Content-Type": annotation.mimetype}
        return Response(content, headers=headers)



# -*- coding: utf-8 -*-
from copy import deepcopy
from MyCapytain.common.reference import URN
import hashlib


class Target(object):
    """ Object and prototype for representing target of annotation.

    .. note:: Target default object are URN based because that's what Nemo is about.

    :param urn: URN targeted by an Annotation
    :type urn: MyCapytain.common.reference.URN

    :ivar urn: Target urn
    """

    def __init__(self, objectId, subreference=None, **kwargs):
        if isinstance(objectId, URN):
            if objectId.reference is not None:
                subreference = str(objectId.reference)
                objectId = str(objectId.upTo(URN.VERSION))
            else:
                objectId = str(objectId)
                subreference = None
        elif isinstance(objectId, tuple):
            objectId, subreference = objectId
        self.__objectId__ = objectId
        self.__subreference__ = subreference

    @property
    def objectId(self):
        return self.__objectId__

    @property
    def subreference(self):
        return self.__subreference__

    def to_json(self):
        """ Method to call to get a serializable object for json.dump or jsonify based on the target

        :return: dict
        """
        if self.subreference is not None:
            return {
                "source": self.objectId,
                "selector": {
                    "type": "FragmentSelector",
                    "conformsTo": "http://ontology-dts.org/terms/subreference",
                    "value": self.subreference
                }
            }
        else:
            return {"source": self.objectId}


class AnnotationResource(object):
    """ Object representing an annotation. It encapsulates both the body (through the .read() function) and the target (through the .target method)

    :param uri: URI identifier for the AnnotationResource
    :type uri: str
    :param target: the Target of the Annotation
    :type target: Target or str or URN or tuple
    :param type_uri: the URI identifying the underlying datatype of the Annotation
    :type type_uri: str
    :param resolver: Resolver providing access to the annotation
    :type resolver: AnnotationResolver
    :param target_class: Alias for the Target class to be used
    :type target_class: class
    :param mimetype: MimeType of the Annotation object
    :type mimetype: str
    :param slug: Slug type of the object
    :type slug: str

    :ivar mimetype: Mimetype of the annotation object
    :ivar sha: SHA identifying the object
    :ivar uri: Original URI of the object
    :ivar slug: Slug Type of the Annotation Object
    :ivar type_uri: URI of the type
    :ivar expandable: Indication of expandability of the object
    :ivar target: Target object of the Annotation
    """

    SLUG = "annotation"

    def __init__(self, uri, target, type_uri, resolver, target_class=Target, mimetype=None, slug=None, **kwargs):
        self.__uri__ = uri
        if not isinstance(target, Target):
            self.__target__ = target_class(target)
        else:
            self.__target__ = target
        self.__type_uri__ = type_uri
        self.__slug__ = slug or deepcopy(type(self).SLUG)
        self.__sha__ = hashlib.sha256(
            "{uri}::{type_uri}".format(uri=uri, type_uri=type_uri).encode('utf-8')
        ).hexdigest()

        self.__content__ = None
        self.__resolver__ = resolver
        self.__retriever__ = None
        self.__mimetype__ = mimetype

    @property
    def mimetype(self):
        return self.__mimetype__

    @property
    def sha(self):
        return self.__sha__

    @property
    def uri(self):
        return self.__uri__

    @property
    def type_uri(self):
        return self.__type_uri__

    @property
    def slug(self):
        return self.__slug__

    @property
    def target(self):
        return self.__target__

    @property
    def expandable(self):
        # default AnnotationResource type is not expandable
        return False

    def read(self):
        """ Read the contents of the Annotation Resource

        :return: the contents of the resource
        :rtype: str or bytes or flask.response
        """
        if not self.__content__:
            self.__retriever__ = self.__resolver__.resolve(self.uri)
            self.__content__, self.__mimetype__ = self.__retriever__.read(self.uri)
        return self.__content__
 
    def expand(self): 
        """ Expand the contents of the Annotation if it is expandable  (i.e. if it references  multiple resources)

        :return: the list of expanded resources
        :rtype: list(AnnotationResource)
        """
        # default AnnotationResource type doesn't expand
        return []



# -*- coding: utf-8 -*-

from copy import deepcopy as copy
from flask_nemo.common import resource_qualifier, ASSETS_STRUCTURE


class PluginPrototype(object):
    """ Prototype for Nemo Plugins

    :param name: Name of the instance of the plugins. Defaults to the class name (Default : Plugin's class name)
    :param namespacing: Add namespace to route to avoid overwriting (Default : False)

    :cvar ROUTES: Routes represents the routes to be added to the Nemo instance. They take the form of a 3-tuple such as :code:`("/read/<collection>", "r_collection", ["GET"])`
    :type ROUTES: list
    :cvar TEMPLATES: Dictionaries of template namespaces and directory to retrieve templates in given namespace
    :type TEMPLATES: dict
    :cvar FILTERS: List of filters to register. Naming convention is :code:`f_doSomething`
    :type FILTERS: list
    :cvar HAS_AUGMENT_RENDER: Enables post-processing in view rendering function :code:`Nemo().render(template, **kwargs)`
    :type HAS_AUGMENT_RENDER: bool
    :cvar CLEAR_ROUTES: Removes original nemo routes
    :type CLEAR_ROUTES: bool
    :cvar CLEAR_ASSETS: Removes original nemo secondary assets
    :type CLEAR_ASSETS: bool
    :cvar STATIC_FOLDER: Overwrite Nemo default statics folder
    :type STATIC_FOLDER: str
    :cvar CSS: List of CSS resources to link in and give access to if local
    :type CSS: [css]
    :cvar JS: List of JS resources to link in and give access to if local
    :type JS: [str]
    :cvar STATIC: List of static resources (images for example) to give access to
    :type STATIC: [str]
    :cvar CACHED: List of functions to cache

    :ivar assets: Dictionary of assets, with each key (css, js and static) being list of resources
    :type assets: dict(str:[str])
    :ivar augment: If true, means that the plugin has a render method which needs to be called upon rendering the view
    :type augment: bool
    :ivar clear_routes: If true, means that the plugin requires Nemo original routes to be removed
    :type clear_routes: bool
    :ivar clear_assets: If true, means that the plugin required Nemo original assets to be removed
    :type clear_assets: bool
    :ivar name: Name of the plugin instance
    :type name: str
    :ivar static_folder: Path to the plugin own static_folder to be used instead of the Nemo default one
    :type static_folder: str
    :ivar namespaced: Indicate if the plugin is namespaced or not
    :type namespaced: bool
    :ivar routes: List of routes where the first member is a flask URL template, the second a method name, and the \
        third a list of accepted Methods
    :type routes: [(str, str, [str])]
    :ivar filters: List of filters method names to be registered in Nemo
    :type filters: [str]
    :ivar templates: Dictionary of namespace and target directory to resolve templates name
    :type templates: {str:str}
    :ivar nemo: Nemo instance
    :type nemo: flask_ext.Nemo
    :ivar name: Name of the plugin instance
    :type name: str

    :Example:

    .. code-block:: python

        ROUTES = [
            # (Path like flask, Name of the function (convention is r_*), List of Http Methods)
            ("/read/<collection>/<textgroup>/<work>/<version>/<passage_identifier>/<visavis>", "r_double", ["GET"])
        ]

    """
    ROUTES = []
    TEMPLATES = {}
    FILTERS = []
    HAS_AUGMENT_RENDER = False
    CLEAR_ROUTES = False
    CLEAR_ASSETS = False
    CSS = []
    STATICS = []
    CACHED = []
    JS = []
    STATIC_FOLDER = None

    def __init__(self, name=None, namespacing=False, *args, **kwargs):
        self.__nemo__ = None
        self.__instance_name__ = name
        if not name:
            self.__instance_name__ = type(self).__name__

        self.__clear_routes__ = copy(type(self).CLEAR_ROUTES)
        self.__clear_assets__ = copy(type(self).CLEAR_ASSETS)
        self.__static_folder__ = copy(type(self).STATIC_FOLDER)
        self.__routes__ = copy(type(self).ROUTES)
        self.__filters__ = copy(type(self).FILTERS)
        self.__templates__ = copy(type(self).TEMPLATES)
        self.__augment__ = copy(type(self).HAS_AUGMENT_RENDER)
        self.__namespaced__ = namespacing

        if namespacing:
            for i in range(0, len(self.__routes__)):
                self.__routes__[i] = tuple(
                    ["/{0}{1}".format(self.name, self.__routes__[i][0])] + list(self.__routes__[i])[1:]
                )

            for i in range(0, len(self.__filters__)):
                self.__filters__[i] = "f_{}_{}".format(self.name, self.__filters__[i][2:])

        self.__assets__ = copy(ASSETS_STRUCTURE)
        for css in type(self).CSS:
            key, value = resource_qualifier(css)
            self.__assets__["css"][key] = value
        for js in type(self).JS:
            key, value = resource_qualifier(js)
            self.__assets__["js"][key] = value
        for static in type(self).STATICS:
            key, value = resource_qualifier(static)
            self.__assets__["static"][key] = value

    def register_nemo(self, nemo=None):
        """ Register Nemo on to the plugin instance

        :param nemo: Instance of Nemo
        """
        self.__nemo__ = nemo

    @property
    def assets(self):
        return self.__assets__

    @property
    def augment(self):
        return self.__augment__

    @property
    def clear_routes(self):
        return self.__clear_routes__

    @property
    def clear_assets(self):
        return self.__clear_assets__

    @property
    def name(self):
        return self.__instance_name__

    @property
    def static_folder(self):
        return self.__static_folder__

    @property
    def namespaced(self):
        return self.__namespaced__

    @property
    def routes(self):
        return self.__routes__

    @property
    def filters(self):
        return self.__filters__

    @property
    def templates(self):
        return self.__templates__

    @property
    def nemo(self):
        return self.__nemo__

    def render(self, **kwargs):
        """ View Rendering function that gets triggered before nemo renders the resources and adds informations to \
        pass to the templates

        :param kwargs: Dictionary of arguments to pass to the template
        :return: Dictionary of arguments to pass to the template
        """
        return kwargs




# -*- coding: utf-8 -*-
"""
    Capitains Nemo
    ====

    Extensions for Flask to propose a Nemo extensions
"""

from urllib.parse import quote
import os.path as op
from operator import itemgetter
import jinja2
from flask import render_template, Blueprint, abort, Markup, send_from_directory, Flask, url_for, redirect, request
from lxml import etree
from copy import deepcopy as copy
from pkg_resources import resource_filename
from collections import Callable, OrderedDict
from MyCapytain.common.constants import Mimetypes
from MyCapytain.resources.prototypes.metadata import ResourceCollection
import inspect

import flask_nemo._data
import flask_nemo.filters
from flask_nemo.chunker import level_grouper as __level_grouper__
from flask_nemo.plugins.default import Breadcrumb
from flask_nemo.common import resource_qualifier, ASSETS_STRUCTURE
from flask_nemo.jinjaext import FakeCacheExtension


class Nemo(object):
    """ Nemo is an extension for Flask python micro-framework which provides
    a User Interface to your app for dealing with CTS API.

    :param app: Flask application
    :type app: Flask
    :param resolver: MyCapytain resolver
    :type resolver: MyCapytain.resolvers.prototypes.Resolver
    :param base_url: Base URL to use when registering the endpoint
    :type base_url: str
    :param cache: Flask-Caching instance or any object having a memoize decorator
    :type cache: flask_caching.Cache
    :param plugins: List of plugins to connect to the Nemo instance
    :type plugins: list(flask_nemo.plugin.PluginPrototype)
    :param template_folder: Folder in which the full set of main namespace templates can be found
    :type template_folder: str
    :param static_folder: Folder in which statics file can be found
    :type static_folder: str
    :param static_url_path: Base url to use for assets
    :type static_url_path: str
    :param urls: Function and routes to register (See Nemo.ROUTES)
    :type urls: [(str, str, [str])]
    :param transform: Dictionary of XSL filepath or transform function where default key is the default applied function
    :type transform: bool|dict
    :param chunker: Dictionary of function to group responses of GetValidReff
    :type chunker: {str: function(str, function(int))}
    :param css: Path to additional stylesheets to load
    :type css: [str]
    :param js: Path to additional javascripts to load
    :type js: [str]
    :param templates: Register or override templates (Dictionary of namespace / directory containing template)
    :type templates: {str: str}
    :param statics: Path to additional statics such as picture to load
    :type statics: [str]
    :param prevent_plugin_clearing_assets: Prevent plugins to clear the static folder route
    :type prevent_plugin_clearing_assets: bool
    :param original_breadcrumb: Use the default Breadcrumb plugin packaged with Nemo (Default: True)
    :type original_breadcrumb: bool
    :param default_lang: Default lang to fall back to
    :type default_lang: str

    :ivar assets: Dictionary of assets loaded individually
    :ivar plugins: List of loaded plugins
    :ivar resolver: Resolver
    :ivar cached: List of cached functions
    :ivar cache: Cache Instance

    .. warning:: Until a C libxslt error is fixed ( https://bugzilla.gnome.org/show_bug.cgi?id=620102 ), \
    it is not possible to use strip spaces in the xslt given to this application. See :ref:`lxml.strip-spaces`
    """

    ROUTES = [
        ("/", "r_index", ["GET"]),
        ("/collections", "r_collections", ["GET"]),
        ("/collections/<objectId>", "r_collection", ["GET"]),
        ("/text/<objectId>/references", "r_references", ["GET"]),
        ("/text/<objectId>/passage/<subreference>", "r_passage", ["GET"]),
        ("/text/<objectId>/passage", "r_first_passage", ["GET"])
    ]
    SEMANTIC_ROUTES = [
        "r_collection", "r_references", "r_passage"
    ]
    FILTERS = [
        "f_formatting_passage_reference",
        "f_i18n_iso",
        "f_order_resource_by_lang",
        "f_hierarchical_passages",
        "f_is_str",
        "f_i18n_citation_type",
        "f_slugify"
    ]

    CACHED = [
        # Routes
        "r_index", "r_collection", "r_collections", "r_references", "r_passage", "r_first_passage", "r_assets",
        # Controllers
        "get_inventory", "get_collection", "get_reffs", "get_passage", "get_siblings",
        # Translater
        "semantic", "make_coins", "expose_ancestors_or_children", "make_members", "transform",
        # Business logic
        # "view_maker", "route", #"render",
    ]

    """ Assets dictionary model
    """
    ASSETS = copy(ASSETS_STRUCTURE)
    default_chunker = __level_grouper__

    def __init__(self,
                 name=None, app=None, base_url="/nemo",
                 cache=None, resolver=None,
                 plugins=None,
                 template_folder=None, static_folder=None, static_url_path=None,
                 urls=None, transform=None, chunker=None, prevnext=None,
                 css=None, js=None, templates=None, statics=None,
                 prevent_plugin_clearing_assets=False,
                 original_breadcrumb=True, default_lang="eng"):

        self.name = __name__
        if name:
            self.name = name
        self.prefix = base_url

        self.resolver = resolver

        if app is not None:
            self.app = app
        else:
            self.app = None

        self.cache = cache
        self.cached = list()
        for func in self.CACHED:
            self.cached.append((getattr(self, func), self))

        self.prevent_plugin_clearing_assets = prevent_plugin_clearing_assets

        if template_folder:
            self.template_folder = template_folder
        else:
            self.template_folder = resource_filename("flask_nemo", "data/templates")

        if static_folder:
            self.static_folder = static_folder
        else:
            self.static_folder = resource_filename("flask_nemo", "data/static")

        if static_url_path:
            self.static_url_path = static_url_path
        else:
            self.static_url_path = "/assets/nemo"
        self.blueprint = None

        if urls:
            self._urls = urls
        else:
            self._urls = copy(type(self).ROUTES)

        # Adding instance information
        self._urls = [tuple(list(url) + [None]) for url in self._urls]
        # Adding semantic ones
        self._semantic_url = [
            (uri+"/<semantic>", endpoint, method, instance)
            for uri, endpoint, method, instance in self._urls
            if endpoint in self.SEMANTIC_ROUTES
        ]

        self._filters = copy(Nemo.FILTERS)
        self._filters = [tuple([filt] + [None]) for filt in self._filters]

        # Reusing self._inventory across requests
        self._inventory = None
        self.__transform = {
            "default": None
        }

        self.__urntransform = {
            "default": None
        }

        if isinstance(transform, dict):
            self.__transform.update(transform)

        self.chunker = {
            "default": type(self).default_chunker
        }
        if isinstance(chunker, dict):
            self.chunker.update(chunker)

        # Setting up assets
        self.__assets__ = copy(type(self).ASSETS)
        if css and isinstance(css, list):
            for css_s in css:
                filename, directory = resource_qualifier(css_s)
                self.__assets__["css"][filename] = directory
        if js and isinstance(js, list):
            for javascript in js:
                filename, directory = resource_qualifier(javascript)
                self.__assets__["js"][filename] = directory
        if statics and isinstance(statics, list):
            for static in statics:
                directory, filename = op.split(static)
                self.__assets__["static"][filename] = directory

        self.__plugins_render_views__ = []
        self.__plugins__ = OrderedDict()
        if original_breadcrumb:
            self.__plugins__["nemo.breadcrumb"] = Breadcrumb(name="breadcrumb")
        if isinstance(plugins, list):
            for plugin in plugins:
                self.__plugins__[plugin.name] = plugin

        self.__templates_namespaces__ = [
            ("main", self.template_folder)
        ]
        self.__instance_templates__ = []

        if isinstance(templates, dict):
            self.__instance_templates__.extend(
                [(namespace, folder) for namespace, folder in templates.items()]
            )
        self.__template_loader__ = dict()
        self.__default_lang__ = default_lang

        if app:
            self.init_app(self.app)

    @property
    def plugins(self):
        """ Dictionary of registered plugins

        :rtype: dict
        """
        return self.__plugins__

    @property
    def assets(self):
        """ Dictionary of assets (First level : type, second level resource)

        :rtype: dict
        """
        return self.__assets__

    @property
    def inventory(self):
        """ Root collection of the application

        :rtype: Collection
        """
        return self.get_inventory()

    def init_app(self, app=None):
        """ Initiate the application

        :param app: Flask application on which to add the extension
        :type app: flask.Flask
        """
        # Legacy code
        if app:
            self.app = app

        self.register()

    def get_locale(self):
        """ Retrieve the best matching locale using request headers

        .. note:: Probably one of the thing to enhance quickly.

        :rtype: str
        """
        best_match = request.accept_languages.best_match(['de', 'fr', 'en', 'la'])
        if best_match is None:
            if len(request.accept_languages) > 0:
                best_match = request.accept_languages[0][0][:2]
            else:
                return self.__default_lang__
        lang = self.__default_lang__
        if best_match == "de":
            lang = "ger"
        elif best_match == "fr":
            lang = "fre"
        elif best_match == "en":
            lang = "eng"
        elif best_match == "la":
            lang = "lat"
        return lang

    def transform(self, work, xml, objectId, subreference=None):
        """ Transform input according to potentially registered XSLT

        .. note:: Since 1.0.0, transform takes an objectId parameter which represent the passage which is called

        .. note:: Due to XSLT not being able to be used twice, we rexsltise the xml at every call of xslt

        .. warning:: Until a C libxslt error is fixed ( https://bugzilla.gnome.org/show_bug.cgi?id=620102 ), \
        it is not possible to use strip tags in the xslt given to this application

        :param work: Work object containing metadata about the xml
        :type work: MyCapytains.resources.inventory.Text
        :param xml: XML to transform
        :type xml: etree._Element
        :param objectId: Object Identifier
        :type objectId: str
        :param subreference: Subreference
        :type subreference: str
        :return: String representation of transformed resource
        :rtype: str
        """
        # We check first that we don't have
        if str(objectId) in self.__transform:
            func = self.__transform[str(objectId)]
        else:
            func = self.__transform["default"]

        # If we have a string, it means we get a XSL filepath
        if isinstance(func, str):
            with open(func) as f:
                xslt = etree.XSLT(etree.parse(f))
            return etree.tostring(
                xslt(xml),
                encoding=str, method="html",
                xml_declaration=None, pretty_print=False, with_tail=True, standalone=None
            )

        # If we have a function, it means we return the result of the function
        elif isinstance(func, Callable):
            return func(work, xml, objectId, subreference)
        # If we have None, it means we just give back the xml
        elif func is None:
            return etree.tostring(xml, encoding=str)

    def get_inventory(self):
        """ Request the api endpoint to retrieve information about the inventory

        :return: Main Collection
        :rtype: Collection
        """
        if self._inventory is not None:
            return self._inventory

        self._inventory = self.resolver.getMetadata()
        return self._inventory

    def get_collection(self, objectId):
        """ Retrieve a collection in the inventory

        :param objectId: Collection Identifier
        :type objectId: str
        :return: Requested collection
        :rtype: Collection
        """
        return self.inventory[objectId]

    def get_reffs(self, objectId, subreference=None, collection=None, export_collection=False):
        """ Retrieve and transform a list of references.

        Returns the inventory collection object with its metadata and a callback function taking a level parameter \
        and returning a list of strings.

        :param objectId: Collection Identifier
        :type objectId: str
        :param subreference: Subreference from which to retrieve children
        :type subreference: str
        :param collection: Collection object bearing metadata
        :type collection: Collection
        :param export_collection: Return collection metadata
        :type export_collection: bool
        :return: Returns either the list of references, or the text collection object with its references as tuple
        :rtype: (Collection, [str]) or [str]
        """
        if collection is not None:
            text = collection
        else:
            text = self.get_collection(objectId)
        reffs = self.chunk(
            text,
            lambda level: self.resolver.getReffs(objectId, level=level, subreference=subreference)
        )
        if export_collection is True:
            return text, reffs
        return reffs

    def get_passage(self, objectId, subreference):
        """ Retrieve the passage identified by the parameters

        :param objectId: Collection Identifier
        :type objectId: str
        :param subreference: Subreference of the passage
        :type subreference: str
        :return: An object bearing metadata and its text
        :rtype: InteractiveTextualNode
        """
        passage = self.resolver.getTextualNode(
            textId=objectId,
            subreference=subreference,
            metadata=True
        )
        return passage

    def get_siblings(self, objectId, subreference, passage):
        """ Get siblings of a browsed subreference

        .. note:: Since 1.0.0c, there is no more prevnext dict. Nemo uses the list of original\
        chunked references to retrieve next and previous, or simply relies on the resolver to get siblings\
        when the subreference is not found in given original chunks.

        :param objectId: Id of the object
        :param subreference: Subreference of the object
        :param passage: Current Passage
        :return: Previous and next references
        :rtype: (str, str)
        """
        reffs = [reff for reff, _ in self.get_reffs(objectId)]
        if subreference in reffs:
            index = reffs.index(subreference)
            # Not the first item and not the last one
            if 0 < index < len(reffs) - 1:
                return reffs[index-1], reffs[index+1]
            elif index == 0 and index < len(reffs) - 1:
                return None, reffs[1]
            elif index > 0 and index == len(reffs) - 1:
                return reffs[index-1], None
        else:
            return passage.siblingsId

    def semantic(self, collection, parent=None):
        """ Generates a SEO friendly string for given collection

        :param collection: Collection object to generate string for
        :param parent: Current collection parent
        :return: SEO/URL Friendly string
        """
        if parent is not None:
            collections = parent.parents[::-1] + [parent, collection]
        else:
            collections = collection.parents[::-1] + [collection]

        return filters.slugify("--".join([item.get_label() for item in collections if item.get_label()]))

    def make_coins(self, collection, text, subreference="", lang=None):
        """ Creates a CoINS Title string from information

        :param collection: Collection to create coins from
        :param text: Text/Passage object
        :param subreference: Subreference
        :param lang: Locale information
        :return: Coins HTML title value
        """
        if lang is None:
            lang = self.__default_lang__
        return "url_ver=Z39.88-2004"\
                 "&ctx_ver=Z39.88-2004"\
                 "&rft_val_fmt=info%3Aofi%2Ffmt%3Akev%3Amtx%3Abook"\
                 "&rft_id={cid}"\
                 "&rft.genre=bookitem"\
                 "&rft.btitle={title}"\
                 "&rft.edition={edition}"\
                 "&rft.au={author}"\
                 "&rft.atitle={pages}"\
                 "&rft.language={language}"\
                 "&rft.pages={pages}".format(
                    title=quote(str(text.get_title(lang))), author=quote(str(text.get_creator(lang))),
                    cid=url_for(".r_collection", objectId=collection.id, _external=True),
                    language=collection.lang, pages=quote(subreference), edition=quote(str(text.get_description(lang)))
                 )

    def expose_ancestors_or_children(self, member, collection, lang=None):
        """ Build an ancestor or descendant dict view based on selected information

        :param member: Current Member to build for
        :param collection: Collection from which we retrieved it
        :param lang: Language to express data in
        :return:
        """
        x = {
            "id": member.id,
            "label": str(member.get_label(lang)),
            "model": str(member.model),
            "type": str(member.type),
            "size": member.size,
            "semantic": self.semantic(member, parent=collection)
        }
        if isinstance(member, ResourceCollection):
            x["lang"] = str(member.lang)
        return x

    def make_members(self, collection, lang=None):
        """ Build member list for given collection

        :param collection: Collection to build dict view of for its members
        :param lang: Language to express data in
        :return: List of basic objects
        """
        objects = sorted([
                self.expose_ancestors_or_children(member, collection, lang=lang)
                for member in collection.members
                if member.get_label()
            ],
            key=itemgetter("label")
        )
        return objects

    def make_parents(self, collection, lang=None):
        """ Build parents list for given collection

        :param collection: Collection to build dict view of for its members
        :param lang: Language to express data in
        :return: List of basic objects
        """
        return [
            {
                "id": member.id,
                "label": str(member.get_label(lang)),
                "model": str(member.model),
                "type": str(member.type),
                "size": member.size
            }
            for member in collection.parents
            if member.get_label()
        ]

    def r_index(self):
        """ Homepage route function

        :return: Template to use for Home page
        :rtype: {str: str}
        """
        return {"template": "main::index.html"}

    def r_collections(self, lang=None):
        """ Retrieve the top collections of the inventory

        :param lang: Lang in which to express main data
        :type lang: str
        :return: Collections information and template
        :rtype: {str: Any}
        """
        collection = self.resolver.getMetadata()
        return {
            "template": "main::collection.html",
            "current_label": collection.get_label(lang),
            "collections": {
                "members": self.make_members(collection, lang=lang)
            }
        }

    def r_collection(self, objectId, lang=None):
        """ Collection content browsing route function

        :param objectId: Collection identifier
        :type objectId: str
        :param lang: Lang in which to express main data
        :type lang: str
        :return: Template and collections contained in given collection
        :rtype: {str: Any}
        """
        collection = self.resolver.getMetadata(objectId)
        return {
            "template": "main::collection.html",
            "collections": {
                "current": {
                    "label": str(collection.get_label(lang)),
                    "id": collection.id,
                    "model": str(collection.model),
                    "type": str(collection.type),
                },
                "members": self.make_members(collection, lang=lang),
                "parents": self.make_parents(collection, lang=lang)
            },
        }

    def r_references(self, objectId, lang=None):
        """ Text exemplar references browsing route function

        :param objectId: Collection identifier
        :type objectId: str
        :param lang: Lang in which to express main data
        :type lang: str
        :return: Template and required information about text with its references
        """
        collection, reffs = self.get_reffs(objectId=objectId, export_collection=True)
        return {
            "template": "main::references.html",
            "objectId": objectId,
            "citation": collection.citation,
            "collections": {
                "current": {
                    "label": collection.get_label(lang),
                    "id": collection.id,
                    "model": str(collection.model),
                    "type": str(collection.type),
                },
                "parents": self.make_parents(collection, lang=lang)
            },
            "reffs": reffs
        }

    def r_first_passage(self, objectId):
        """ Provides a redirect to the first passage of given objectId

        :param objectId: Collection identifier
        :type objectId: str
        :return: Redirection to the first passage of given text
        """
        collection, reffs = self.get_reffs(objectId=objectId, export_collection=True)
        first, _ = reffs[0]
        return redirect(
            url_for(".r_passage_semantic", objectId=objectId, subreference=first, semantic=self.semantic(collection))
        )

    def r_passage(self, objectId, subreference, lang=None):
        """ Retrieve the text of the passage

        :param objectId: Collection identifier
        :type objectId: str
        :param lang: Lang in which to express main data
        :type lang: str
        :param subreference: Reference identifier
        :type subreference: str
        :return: Template, collections metadata and Markup object representing the text
        :rtype: {str: Any}
        """
        text = self.get_passage(objectId=objectId, subreference=subreference)
        collection = self.get_collection(objectId)
        passage = self.transform(text, text.export(Mimetypes.PYTHON.ETREE), objectId)
        prev, next = self.get_siblings(objectId, subreference, text)
        return {
            "template": "main::text.html",
            "objectId": objectId,
            "subreference": subreference,
            "collections": {
                "current": {
                    "label": collection.get_label(lang),
                    "id": collection.id,
                    "model": str(collection.model),
                    "type": str(collection.type),
                    "author": text.get_creator(lang),
                    "title": text.get_title(lang),
                    "description": text.get_description(lang),
                    "citation": collection.citation,
                    "coins": self.make_coins(collection, text, subreference, lang=lang)
                },
                "parents": self.make_parents(collection, lang=lang)
            },
            "text_passage": Markup(passage),
            "prev": prev,
            "next": next
        }

    def r_assets(self, filetype, asset):
        """ Route for specific assets.

        :param filetype: Asset Type
        :param asset: Filename of an asset
        :return: Response
        """
        if filetype in self.assets and asset in self.assets[filetype] and self.assets[filetype][asset]:
            return send_from_directory(
                directory=self.assets[filetype][asset],
                filename=asset
            )
        abort(404)

    def register_assets(self):
        """ Merge and register assets, both as routes and dictionary

        :return: None
        """
        self.blueprint.add_url_rule(
            # Register another path to ensure assets compatibility
            "{0}.secondary/<filetype>/<asset>".format(self.static_url_path),
            view_func=self.r_assets,
            endpoint="secondary_assets",
            methods=["GET"]
        )

    def create_blueprint(self):
        """ Create blueprint and register rules

        :return: Blueprint of the current nemo app
        :rtype: flask.Blueprint
        """
        self.register_plugins()

        self.blueprint = Blueprint(
            self.name,
            "nemo",
            url_prefix=self.prefix,
            template_folder=self.template_folder,
            static_folder=self.static_folder,
            static_url_path=self.static_url_path
        )

        for url, name, methods, instance in self._urls:
            self.blueprint.add_url_rule(
                url,
                view_func=self.view_maker(name, instance),
                endpoint=_plugin_endpoint_rename(name, instance),
                methods=methods
            )

        for url, name, methods, instance in self._semantic_url:
            self.blueprint.add_url_rule(
                url,
                view_func=self.view_maker(name, instance),
                endpoint=_plugin_endpoint_rename(name, instance)+"_semantic",
                methods=methods
            )

        self.register_assets()
        self.register_filters()

        # We extend the loading list by the instance value
        self.__templates_namespaces__.extend(self.__instance_templates__)
        # We generate a template loader
        for namespace, directory in self.__templates_namespaces__[::-1]:
            if namespace not in self.__template_loader__:
                self.__template_loader__[namespace] = []
            self.__template_loader__[namespace].append(
                jinja2.FileSystemLoader(op.abspath(directory))
            )
        self.blueprint.jinja_loader = jinja2.PrefixLoader(
            {namespace: jinja2.ChoiceLoader(paths) for namespace, paths in self.__template_loader__.items()},
            "::"
        )

        if self.cache is not None:
            for func, instance in self.cached:
                setattr(instance, func.__name__, self.cache.memoize()(func))

        return self.blueprint

    def view_maker(self, name, instance=None):
        """ Create a view

        :param name: Name of the route function to use for the view.
        :type name: str
        :return: Route function which makes use of Nemo context (such as menu informations)
        :rtype: function
        """
        if instance is None:
            instance = self
        sig = "lang" in [
            parameter.name
            for parameter in inspect.signature(getattr(instance, name)).parameters.values()
        ]

        def route(**kwargs):
            if sig and "lang" not in kwargs:
                kwargs["lang"] = self.get_locale()
            if "semantic" in kwargs:
                del kwargs["semantic"]
            return self.route(getattr(instance, name), **kwargs)
        return route

    def main_collections(self, lang=None):
        """ Retrieve main parent collections of a repository

        :param lang: Language to retrieve information in
        :return: Sorted collections representations
        """
        return sorted([
            {
                "id": member.id,
                "label": str(member.get_label(lang=lang)),
                "model": str(member.model),
                "type": str(member.type),
                "size": member.size
            }
            for member in self.resolver.getMetadata().members
        ], key=itemgetter("label"))

    def make_cache_keys(self, endpoint, kwargs):
        """ This function is built to provide cache keys for templates

        :param endpoint: Current endpoint
        :param kwargs: Keyword Arguments
        :return: tuple of i18n dependant cache key and i18n ignoring cache key
        :rtype: tuple(str)
        """
        keys = sorted(kwargs.keys())
        i18n_cache_key = endpoint+"|"+"|".join([kwargs[k] for k in keys])
        if "lang" in keys:
            cache_key = endpoint+"|" + "|".join([kwargs[k] for k in keys if k != "lang"])
        else:
            cache_key = i18n_cache_key
        return i18n_cache_key, cache_key

    def render(self, template, **kwargs):
        """ Render a route template and adds information to this route.

        :param template: Template name.
        :type template: str
        :param kwargs: dictionary of named arguments used to be passed to the template
        :type kwargs: dict
        :return: Http Response with rendered template
        :rtype: flask.Response
        """

        kwargs["cache_key"] = "%s" % kwargs["url"].values()
        kwargs["lang"] = self.get_locale()
        kwargs["assets"] = self.assets
        kwargs["main_collections"] = self.main_collections(kwargs["lang"])
        kwargs["cache_active"] = self.cache is not None
        kwargs["cache_time"] = 0
        kwargs["cache_key"], kwargs["cache_key_i18n"] = self.make_cache_keys(request.endpoint, kwargs["url"])
        kwargs["template"] = template

        for plugin in self.__plugins_render_views__:
            kwargs.update(plugin.render(**kwargs))

        return render_template(kwargs["template"], **kwargs)

    def route(self, fn, **kwargs):
        """ Route helper : apply fn function but keep the calling object, *ie* kwargs, for other functions

        :param fn: Function to run the route with
        :type fn: function
        :param kwargs: Parsed url arguments
        :type kwargs: dict
        :return: HTTP Response with rendered template
        :rtype: flask.Response
        """
        new_kwargs = fn(**kwargs)

        # If there is no templates, we assume that the response is finalized :
        if not isinstance(new_kwargs, dict):
            return new_kwargs

        new_kwargs["url"] = kwargs
        return self.render(**new_kwargs)

    def register(self):
        """ Register the app using Blueprint

        :return: Nemo blueprint
        :rtype: flask.Blueprint
        """
        if self.app is not None:
            if not self.blueprint:
                self.blueprint = self.create_blueprint()
            self.app.register_blueprint(self.blueprint)
            if self.cache is None:
                # We register a fake cache extension.
                setattr(self.app.jinja_env, "_fake_cache_extension", self)
                self.app.jinja_env.add_extension(FakeCacheExtension)
            return self.blueprint
        return None

    def register_filters(self):
        """ Register filters for Jinja to use

       .. note::  Extends the dictionary filters of jinja_env using self._filters list
        """
        for _filter, instance in self._filters:
            if not instance:
                self.app.jinja_env.filters[
                    _filter.replace("f_", "")
                ] = getattr(flask_nemo.filters, _filter)
            else:
                self.app.jinja_env.filters[
                    _filter.replace("f_", "")
                ] = getattr(instance, _filter.replace("_{}".format(instance.name), ""))

    def register_plugins(self):
        """ Register plugins in Nemo instance

        - Clear routes first if asked by one plugin
        - Clear assets if asked by one plugin and replace by the last plugin registered static_folder
        - Register each plugin
            - Append plugin routes to registered routes
            - Append plugin filters to registered filters
            - Append templates directory to given namespaces
            - Append assets (CSS, JS, statics) to given resources 
            - Append render view (if exists) to Nemo.render stack
        """
        if len([plugin for plugin in self.__plugins__.values() if plugin.clear_routes]) > 0:  # Clear current routes
            self._urls = list()
            self.cached = list()

        clear_assets = [plugin for plugin in self.__plugins__.values() if plugin.clear_assets]
        if len(clear_assets) > 0 and not self.prevent_plugin_clearing_assets:  # Clear current Assets
            self.__assets__ = copy(type(self).ASSETS)
            static_path = [plugin.static_folder for plugin in clear_assets if plugin.static_folder]
            if len(static_path) > 0:
                self.static_folder = static_path[-1]

        for plugin in self.__plugins__.values():
            self._urls.extend([(url, function, methods, plugin) for url, function, methods in plugin.routes])
            self._filters.extend([(filt, plugin) for filt in plugin.filters])
            self.__templates_namespaces__.extend(
                [(namespace, directory) for namespace, directory in plugin.templates.items()]
            )
            for asset_type in self.__assets__:
                for key, value in plugin.assets[asset_type].items():
                    self.__assets__[asset_type][key] = value
            if plugin.augment:
                self.__plugins_render_views__.append(plugin)

            if hasattr(plugin, "CACHED"):
                for func in plugin.CACHED:
                    self.cached.append((getattr(plugin, func), plugin))
            plugin.register_nemo(self)

    def chunk(self, text, reffs):
        """ Handle a list of references depending on the text identifier using the chunker dictionary.

        :param text: Text object from which comes the references
        :type text: MyCapytains.resources.texts.api.Text
        :param reffs: List of references to transform
        :type reffs: References
        :return: Transformed list of references
        :rtype: [str]
        """
        if str(text.id) in self.chunker:
            return self.chunker[str(text.id)](text, reffs)
        return self.chunker["default"](text, reffs)


def _plugin_endpoint_rename(fn_name, instance):
    """ Rename endpoint function name to avoid conflict when namespacing is set to true

    :param fn_name: Name of the route function
    :param instance: Instance bound to the function
    :return: Name of the new namespaced function name
    """

    if instance and instance.namespaced:
        fn_name = "r_{0}_{1}".format(instance.name, fn_name[2:])
    return fn_name


"""

Test file for filters.

"""

from unittest import TestCase
from flask import Flask
from flask_nemo import Nemo
from flask_nemo.filters import f_formatting_passage_reference, f_annotation_filter,\
    f_hierarchical_passages, f_i18n_citation_type, f_i18n_iso, f_is_str, f_order_resource_by_lang
from MyCapytain.common.reference import Citation
from tests.test_resources import NemoResource


class TestFilters(NemoResource, TestCase):

    def test_f_formatting_passage_reference(self):
        """ Test split of passage range identifier
        """
        self.assertEqual(f_formatting_passage_reference("1.1-1.2"), "1.1")
        self.assertEqual(f_formatting_passage_reference("1.1"), "1.1")

    def test_f_i18n_iso(self):
        """ Test split of passage range identifier
        """
        self.assertEqual(f_i18n_iso("eng"), "English")
        self.assertEqual(f_i18n_iso("eng", "fre"), "anglais")
        self.assertEqual(f_i18n_iso("eng", "ger"), "English")

    def test_register_filter(self):
        app = Flask(__name__)
        self.nemo = Nemo(app=app)
        self.nemo.register_filters()
        self.assertEqual(self.nemo.app.jinja_env.filters["formatting_passage_reference"], f_formatting_passage_reference)
        self.assertEqual(self.nemo.app.jinja_env.filters["order_resource_by_lang"], f_order_resource_by_lang)
        self.assertEqual(self.nemo.app.jinja_env.filters["hierarchical_passages"], f_hierarchical_passages)

    def test_f_hierarchical_passages(self):
        """Test for normal, simple passage hierarchical conversion
        :return:
        """
        reffs = [("1.5.8", "Line 8"), ("1.5.9", "Line 9"), ("1.6.8", "Line 7"), ("2.5.8", "Line 12")]
        citation_line = Citation(name="line")
        citation_poem = Citation(name="poem", child=citation_line)
        citation_book = Citation(name="book", child=citation_poem)
        converted = f_hierarchical_passages(reffs, citation_book)
        self.assertEqual(converted["%book|1%"]["%poem|5%"]["Line 8"], "1.5.8")
        self.assertEqual(converted["%book|1%"]["%poem|5%"]["Line 9"], "1.5.9")
        self.assertEqual(converted["%book|1%"]["%poem|6%"]["Line 7"], "1.6.8")
        self.assertEqual(converted["%book|2%"]["%poem|5%"]["Line 12"], "2.5.8")
        self.assertEqual(len(converted), 2)
        self.assertEqual(len(converted["%book|1%"]), 2)
        self.assertEqual(len(converted["%book|1%"]["%poem|5%"]), 2)
        self.assertEqual(len(converted["%book|1%"]["%poem|6%"]), 1)
        self.assertEqual(len(converted["%book|2%"]), 1)

    def test_f_hierarchical_range_passages(self):
        """Test for range passage hierarchical conversion
        :return:
        """
        reffs = [("1.5.8-1.5.9", "Line 8"), ("1.5.9-1.5.15", "Line 9"), ("1.6.8-2.9.16", "Line 7"), ("2.5.8-16.45.928", "Line 12")]
        citation_line = Citation(name="line")
        citation_poem = Citation(name="poem", child=citation_line)
        citation_book = Citation(name="book", child=citation_poem)
        converted = f_hierarchical_passages(reffs, citation_book)
        self.assertEqual(converted["%book|1%"]["%poem|5%"]["Line 8"], "1.5.8-1.5.9")
        self.assertEqual(converted["%book|1%"]["%poem|5%"]["Line 9"], "1.5.9-1.5.15")
        self.assertEqual(converted["%book|1%"]["%poem|6%"]["Line 7"], "1.6.8-2.9.16")
        self.assertEqual(converted["%book|2%"]["%poem|5%"]["Line 12"], "2.5.8-16.45.928")
        self.assertEqual(len(converted), 2)
        self.assertEqual(len(converted["%book|1%"]), 2)
        self.assertEqual(len(converted["%book|1%"]["%poem|5%"]), 2)
        self.assertEqual(len(converted["%book|1%"]["%poem|6%"]), 1)
        self.assertEqual(len(converted["%book|2%"]), 1)

    def test_f_is_str(self):
        """ Test string
        """
        self.assertEqual(f_is_str("h"), True)
        self.assertEqual(f_is_str([]), False)

    def test_f_i18n_citation_type(self):
        self.assertEqual(f_i18n_citation_type("%book|1%"), "Book 1")

    def test_annotation_filter(self):
        class Annotation(object):
            def __init__(self, type_uri):
                self.type_uri = type_uri

        a = Annotation("1")
        b = Annotation("2")
        c = Annotation("3")
        d = Annotation("2")

        self.assertEqual(
            f_annotation_filter([a, b, c, d], "1", 1), a,
            "We should get only one of type 1"
        )

        self.assertEqual(
            f_annotation_filter([a, b, c, d], "4", 1), None,
            "We should get nothing"
        )

        self.assertEqual(
            f_annotation_filter([a, b, c, d], "2", 2), d,
            "We should get the second item"
        )


"""
    Test for simple init parameters override
"""
from tests.test_resources import NemoResource, NautilusDummy
from flask_nemo import Nemo
from flask import Flask


class TestNemoInit(NemoResource):
    def test_overwrite_urls(self):
        """ Check that routes can be added or modified
        """
        routes = [("/index.html", "r_index", ["GET"])] + Nemo.ROUTES[1:]
        app = Flask(__name__)
        nemo = Nemo(app=app, urls=routes)
        nemo.register()
        self.assertIn("flask_nemo", app.blueprints)

        rules = [(rule.rule, rule.endpoint) for rule in app.url_map.iter_rules()]
        self.assertIn("/nemo/index.html", [rule[0] for rule in rules])
        self.assertNotIn("/nemo/", [rule[0] for rule in rules])

    def test_static_url_path(self):
        """ Test that the extension static url path is changing when set up
        """
        app = Flask(__name__)
        nemo = Nemo(app=app, static_url_path="/assets/nemoOo")
        nemo.register()
        self.assertIn("flask_nemo", app.blueprints, "Nemo is a registered blueprint")

        rules = [(rule.rule, rule.endpoint) for rule in app.url_map.iter_rules()]
        self.assertIn("/nemo/assets/nemoOo/<path:filename>", [rule[0] for rule in rules])
        self.assertIn("/nemo/assets/nemoOo.secondary/<filetype>/<asset>", [rule[0] for rule in rules])

    def test_static_folder(self):
        """ Test the use of static fikder parameter to implement assets customization
        """
        app = Flask(__name__)
        nemo = Nemo(app=app, static_folder="/examples")
        nemo.register()

        self.assertEqual(nemo.static_folder, "/examples")
        self.assertEqual(nemo.blueprint.static_folder, "/examples")

    def test_template_folder(self):
        """ Test the parameter to set up a different template folder
        """
        app = Flask(__name__)
        nemo = Nemo(app=app, template_folder="/examples")
        nemo.register()

        self.assertEqual(nemo.template_folder, "/examples")
        self.assertEqual(nemo.blueprint.template_folder, "/examples")

    def test_other_endpoint(self):
        """ Test when an endpoint is set """
        nemo = Nemo(resolver=NautilusDummy)
        self.assertEqual(nemo.resolver, NautilusDummy, "Endpoint should be set through endpoint parameter")



from MyCapytain.common.reference import URN
from flask_nemo.query.proto import QueryPrototype
from flask_nemo.query.annotation import AnnotationResource
from werkzeug.exceptions import NotFound


class SimpleQuery(QueryPrototype):
    """ Query Interface for hardcoded annotations.

    :param annotations: List of tuple of (CTS URN Targeted, URI of the Annotation, Type of the annotation) or/and AnnotationResources
    :type annotations: [(str, str, str) or AnnotationResource]
    :param resolver: Resolver
    :type resolver: Resolver

    This interface requires to be connected to Nemo upon instantiation to expand annotations :

    >>> nemo = Nemo("/", endpoint="http://cts.perseids.org")
    >>> query = SimpleQuery([...])
    >>> query.process(nemo)

    """

    def __init__(self, annotations, resolver=None):
        super(SimpleQuery, self).__init__(None)
        self.__annotations__ = []
        self.__nemo__ = None
        self.__resolver__ = resolver

        for resource in annotations:
            if isinstance(resource, tuple):
                target, body, type_uri = resource
                self.__annotations__.append(AnnotationResource(
                    body, target, type_uri, self.__resolver__
                ))
            else:
                self.__annotations__.append(resource)

    @property
    def textResolver(self):
        return self.__nemo__.resolver

    def process(self, nemo):
        """ Register nemo and parses annotations

        .. note:: Process parses the annotation and extends informations about the target URNs by retrieving resource in range

        :param nemo: Nemo
        """
        self.__nemo__ = nemo
        for annotation in self.__annotations__:
            annotation.target.expanded = frozenset(
                self.__getinnerreffs__(
                    objectId=annotation.target.objectId,
                    subreference=annotation.target.subreference
                )
            )

    def __get_resource_metadata__(self, objectId):
        """ Return a metadata text object

        :param objectId: objectId of the text
        :return: Text
        """
        return self.textResolver.getMetadata(objectId)

    @property
    def annotations(self):
        return self.__annotations__

    def getResource(self, sha):
        try:
            return [annotation for annotation in self.annotations if annotation.sha == sha][0]
        except IndexError:
            raise NotFound

    def getAnnotations(self, targets, wildcard=".", include=None, exclude=None, limit=None, start=1, expand=False,
                       **kwargs):
        annotations = []

        if not targets:
            return len(self.annotations), sorted(self.annotations, key=lambda x: x.uri)

        if not isinstance(targets, list):
            targets = [targets]

        for target in targets:
            if isinstance(target, tuple):
                objectId, subreference = target
            elif isinstance(target, URN):
                objectId, subreference = target.upTo(URN.NO_PASSAGE), str(target.reference)
            else:
                objectId, subreference = target, None

            ref_in_range = list(self.__getinnerreffs__(
                objectId=objectId,
                subreference=subreference
            ))
            ref_in_range = frozenset(ref_in_range)
            annotations.extend([
                annotation
                for annotation in self.annotations
                # Exact match
                if (str(objectId), subreference) == (annotation.target.objectId, annotation.target.subreference) or
                # Deeper match
                len(ref_in_range.intersection(annotation.target.expanded)) > 0
            ])

        annotations = list(set(annotations))

        return len(annotations), sorted(annotations, key=lambda x: x.uri)

    def __getinnerreffs__(self, objectId, subreference):
        """ Resolve the list of urns between in a range

        :param text_metadata: Resource Metadata
        :param objectId: ID of the Text
        :type objectId: str
        :param subreference: ID of the Text
        :type subreference: str
        :return: References in the span
        """
        level = 0
        yield subreference
        while level > -1:
            reffs = self.__nemo__.resolver.getReffs(
                objectId,
                subreference=subreference,
                level=level
            )
            if len(reffs) == 0:
                break
            else:
                for r in reffs:
                    yield r
                level += 1



from unittest import TestCase
from flask_nemo.query.proto import QueryPrototype
from flask_nemo.query.annotation import Target, AnnotationResource
from flask_nemo.query.resolve import RetrieverPrototype, Resolver, UnresolvableURIError
import MyCapytain.common.reference


class TestQuery(TestCase):
    """ Test query
    """

    def setUp(self):
        self.fakegetreffs = lambda x:x
        self.retrieverproto = RetrieverPrototype()
        self.resolver = Resolver(self.retrieverproto)
        self.fakeurn = MyCapytain.common.reference.URN("urn:cts:greekLit:tlg0012.tlg001.perseus-grc1")
        self.faketarget = self.fakeurn

    def test_query_prototype(self):
        """ Test that the query prototype returns nothing """
        self.assertEqual((0,[]), QueryPrototype(self.fakegetreffs).getAnnotations(self.fakeurn))

    def test_query_prototype_get_resource(self):
        """ Test that the query prototype returns nothing """
        self.assertEqual(None, QueryPrototype(self.fakegetreffs).getResource("any"))

    def test_retriever_prototype_read(self):
        """ Test that the retriever prototype reads nothing """
        self.assertEqual(
            (None, "text/xml"), self.retrieverproto.read("http://example.org"),
            "Default response when there is nothing is a None object for prototype"
        )

    def test_retriever_prototype_match(self):
        """ Test that the retriever prototype matches nothing """
        self.assertEqual(False, RetrieverPrototype().match("http://example.org"))

    def test_resolver(self):
        """ Test that the default resolver, configured with retriever prototype, raises an exception on resolve """
        with self.assertRaises(UnresolvableURIError, msg="Default resolver should raise exception"):
            self.resolver.resolve("http://example.org")
        
    def test_target_urn(self):
        """ Test that a target returns its urn property """
        self.assertEqual(str(self.fakeurn), Target(self.fakeurn).objectId)

    def test_target_json(self):
        """Default target have a string as json representation"""
        self.assertEqual(
            Target(
                "urn:cts:latinLit:phi1294.phi002.perseus-lat2", "1.2.3"
            ).to_json(),
            {
                "source": "urn:cts:latinLit:phi1294.phi002.perseus-lat2",
                "selector": {
                    "type": "FragmentSelector",
                    "conformsTo": "http://ontology-dts.org/terms/subreference",
                    "value": "1.2.3"
                }
            },
            "Default target have a string as json representation"
        )

    def test_annotation_resource_read(self):
        """ Test that an annotation resource with default resolver raises an exception on read """
        annotation = AnnotationResource(
            "http://example.org/annotation",
            self.faketarget,
            "http://data.perseus.org/rdfvocab/fake",
            self.resolver
        )
        with self.assertRaises(UnresolvableURIError):
            annotation.read()

    def test_annotation_resource_expand(self):
        """ Test that an annotation resource expands to empty list """
        annotation = AnnotationResource(
            "http://example.org/annotation",
            self.faketarget,
            "http://data.perseus.org/rdfvocab/fake",
            self.resolver
        )
        self.assertEqual(
            [], annotation.expand(),
            "Expand should return a list"
        )

    def test_annotation_resource_expandable_property(self):
        """ Test that an annotation resource is not expandable """
        annotation = AnnotationResource(
            "http://example.org/annotation",
            self.faketarget,
            "http://data.perseus.org/rdfvocab/fake",
            self.resolver
        )
        self.assertFalse(annotation.expandable, msg="Default annotation should not be expandable")

    def test_annotation_resource_uri_property(self):
        """ Test that an annotation resource returns its uri """

        annotation = AnnotationResource(
            "http://example.org/annotation",
            self.faketarget,
            "http://data.perseus.org/rdfvocab/fake",
            self.resolver
        )
        self.assertEqual(
            "http://example.org/annotation", annotation.uri,
            "Annotation uri property should read correctly"
        )

    def test_annotation_resource_type_uri_property(self):
        """ Test that an annotation resource returns its uri type"""
        annotation = AnnotationResource(
            "http://example.org/annotation",
            self.faketarget,
            "http://data.perseus.org/rdfvocab/fake",
            self.resolver
        )
        self.assertEqual(
            "http://data.perseus.org/rdfvocab/fake", annotation.type_uri,
            "Annotation type_uri should read correctly"
        )

    def test_annotation_resource_sha_property(self):
        """ Test that an annotation resource returns its uri type"""
        annotation = AnnotationResource(
            "http://example.org/annotation",
            self.faketarget,
            "http://data.perseus.org/rdfvocab/fake",
            self.resolver
        )
        self.assertEqual(
            "fb6737b46ba8b9d8cd7d486c348275c07b017bbb748a63c673a9b685acf0a859", annotation.sha,
            "Annotation sha should be stable"
        )

    def test_annotation_resource_slug_property(self):
        """ Test that an annotation resource returns its slug"""
        annotation = AnnotationResource(
            "http://example.org/annotation",
            self.faketarget,
            "http://data.perseus.org/rdfvocab/fake",
            self.resolver
        )
        self.assertEqual(
            "annotation", annotation.slug,
            "Annotation slug should read correctly"
        )
